/* SofChat (Web Chat program SHM version)      */
/* Copyright(c) 1996-1997 Y.YONEYAMA           */
/*                                             */
/* chat.c                                      */

int version    = 129;    /* current version    */
int patchlevel =   6;    /* patch level        */
int memversion = 129;    /* memory compatible version */

#define DEFAULT_BG    "white"
#define DEFAULT_TX    "black"
#define DEFAULT_LI    ""
#define DEFAULT_VL    ""
#define DEFAULT_AL    ""
#define DEFAULT_DN    "red"
#define DEFAULT_DT    "gray"

#ifndef DEFAULT_METHOD
/*#define DEFAULT_METHOD "GET"*/
#define DEFAULT_METHOD "POST"
#endif

#ifndef DEFAULT_PERMIT
/*#define DEFAULT_PERMIT 1*/
#define DEFAULT_PERMIT 0
#endif

#define SHIFT_TIME      0
#define TZ_STRING      ""
/*
#define SHIFT_TIME  14*3600
#define TZ_STRING    " JST"
*/

#define COOKIE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <unistd.h>

#include "sofchat.h"
#include "utils.h"
#include "html.h"
#include "messages.h"

/* external functions */
void code();      /* code.c */
void decode();    /* code.c */

#define strncpy jstrncpy  /* utils.c */

#define MAX_USERS      80
#define MAX_LINES     100
#define MAX_BBS        50
#define MAX_DENYHOSTS  10
#define MAX_LINKS      10
#define NEWUSER_LINES  15
#define TIMEOUT        30       /* 30 sec. */
#define BLINKTIME      90
#define AUTOLOGOUT    120
#define LOGOUTWAIT      3
#define SLEEPTIME     600       /* (Zzz) 10 min. */
#define REFUSETIME      1       /* default refuse time */

#define TITLE_LEN     256
#define DEF_MAX_LINES  30

typedef struct {
  int  version;                 /* memory version */
  char memlock;                 /* lock status */
  char title[TITLE_LEN];
  char bgcolor[COLOR_LEN];
  char bgimage[ADDR_LEN];
  char txcolor[COLOR_LEN];
  char licolor[COLOR_LEN];
  char vlcolor[COLOR_LEN];
  char alcolor[COLOR_LEN];
  char dncolor[COLOR_LEN];      /* default name color */
  char dtcolor[COLOR_LEN];      /* date color */ 
  USERINFO user[MAX_USERS];
  TEXT topic;
  TEXT text[MAX_LINES];         /* text */
  int  text_line;               /* current line */
  TEXT bbs[MAX_BBS];            /* bbs */
  int  bbs_line;                /* current bbs line */
  HOSTINFO denyhosts[MAX_DENYHOSTS];
  LINK links[MAX_LINKS];
} CHAT;

CHAT *get_chat();
void release_chat();
void lock_chat();
void unlock_chat();
void reset_chat();

#ifdef EXPORT_USER
void export_users();
void import_users();
#endif
#ifdef EXPORT_BBS
void export_bbss();
void import_bbss();
#endif
#ifdef EXPORT_INI
void export_ini();
void import_ini();
#endif
#ifdef EXPORT_CHAT
void export_chat();
void import_chat();
#endif

void entry_form();
void enter_base();
void input_form();

void printlinks();
void printdebug();
void printchat();
void printtopic();
void printbbs();
void printdate();
void sigalm();
void sigterm();
void killtag();
char *getcgiarg();
int setuserinfo();
void sjis2euc();
void euc2sjis();
void header();
void footer();
void scerr();
char *getparam();
void paramencode();


char *cgibin = CHATCGI;
char *official_url = "http://www.asahi-net.or.jp/~fj9y-ynym/sofchat/index.html";
char cgimtd[5] = DEFAULT_METHOD;

char agent[AGENT_LEN]    = "";
char haddress[HADDR_LEN] = "";
IP   haddressip          = 0;

CHAT *chat = NULL, *local = NULL;
int  outhead = 0;
int  sjis = 0;
int  autoreload = 0;
char *title;
char def_title[TITLE_LEN];
char def_bgcolor[COLOR_LEN];
char def_txcolor[COLOR_LEN];
char def_bgimage[ADDR_LEN];
char def_licolor[COLOR_LEN];
char def_vlcolor[COLOR_LEN];
char def_alcolor[COLOR_LEN];
char def_dncolor[COLOR_LEN];
char def_dtcolor[COLOR_LEN];

/* cgi argument */
int refresh             = 0;
int maxlines            = 0;
char bgcolor[COLOR_LEN] = "";
char txcolor[COLOR_LEN] = "";
char name[NAME_LEN]     = "";
char namecolor[COLOR_LEN] = "";
char textcolor[COLOR_LEN] = "";
char password[PASS_LEN] = "";
char address[ADDR_LEN]  = "";
char referer[ADDR_LEN]  = "";
char mode[10]           = "";
char message[MSG_SIZE]  = "";
char new_password[PASS_LEN] = "";
char lang[10]           = "";
char frame[10]          = "";
char secure[10]         = "";
int room = 0;
int autoclear           = 0;

int  _refresh              = 0;
int  _flags                = 0;
char link_url[MAX_LINKS][ADDR_LEN];
char link_label[MAX_LINKS][LABEL_LEN];
int  link_items = 0;

char cpassword[25]     = "";
char alertmsg[MEMO_SIZE*2] = "";
#ifdef COOKIE
char cookie[256]     = "";
#endif /* COOKIE */

time_t     now;
struct tm *nowtm;

int userid = -1;
int userproperty = 0;
int permission = 0;
int hasalert = 0;
int alertbtn = 0;
int back = 0;
int userreseted = 0;

CHATMODE chat_mode;
BODY body;

void chat_exit(ret)
int ret;
{
  if(chat != NULL)
    release_chat(chat);
  exit(ret);
}

void init_memory(chat)
CHAT *chat;
{
  int i;
  memset((char *)chat, 0, sizeof(CHAT));
  chat->version = memversion;

  strncpy(chat->title, MSG(M_TITLE), TITLE_LEN);
  strncpy(chat->bgcolor, DEFAULT_BG, COLOR_LEN);
  strncpy(chat->txcolor, DEFAULT_TX, COLOR_LEN);
  strncpy(chat->licolor, DEFAULT_LI, COLOR_LEN);
  strncpy(chat->vlcolor, DEFAULT_VL, COLOR_LEN);
  strncpy(chat->alcolor, DEFAULT_AL, COLOR_LEN);
  strncpy(chat->dncolor, DEFAULT_DN, COLOR_LEN);
  strncpy(chat->dtcolor, DEFAULT_DT, COLOR_LEN);
  for(i = 0; i < 2; i++)
  {
    strncpy(chat->links[i].label_ja, def_links[i].label_ja, LABEL_LEN);
    strncpy(chat->links[i].label_en, def_links[i].label_en, LABEL_LEN);
    strncpy(chat->links[i].url, def_links[i].url, ADDR_LEN);
  }
  strncpy(chat->text[0].name, MSG(M_TITLE), NAME_LEN);
  strncpy(chat->text[0].message, MSG(M_INITIALIZED), MSG_SIZE);
  chat->text[0].atime = now;
  chat->text_line ++;

#ifdef EXPORT_INI
  import_ini(chat);
#endif
#ifdef EXPORT_USER
  import_users(chat);
#endif
#ifdef EXPORT_BBS
  import_bbss(chat);
#endif
#ifdef EXPORT_CHAT
  import_chat(chat);
#endif
}

int isdenyhost(chat, ip)
CHAT *chat;
IP ip;
{
  int i;
  for(i=0;i<MAX_DENYHOSTS;i++)
  {
    if((chat->denyhosts[i].atime >= now) &&
       (compareIP(chat->denyhosts[i].address, ip, 0xffffff00) == 0))
      return(1);
  }
  return(0);
}

int ismanager()
{
  if((strcmp(name, MANAGER) == 0) &&
     (strcmp(password, PASSWORD) == 0))
    return(-1);
  else
    return(0);
}

int get_tagparam(src, len, arg1, arg2, arg3)
char *src;
int len;
char *arg1;
char *arg2;
char *arg3;
{
  int i;
  *arg1 = '\0';
  *arg2 = '\0';
  *arg3 = '\0';

  if(*src != '<') return(0);
  src++;
  /* arg1 */
  for(i=0;i<(len-1);i++)
  {
    if(*src == '>' || *src == '\0' || *src == ' ') break;
    *(arg1++) = *(src++);
  }
  *arg1 = '\0';
  if(*src != ' ') return(1);
  while(*src == ' ') src++;

  /* arg2 */
  for(i=0;i<(len-1);i++)
  {
    if(*src == '>' || *src == '\0' || *src == ' ') break;
    *(arg2++) = *(src++);
  }
  *arg2 = '\0';
  if(*src != ' ') return(2);
  while(*src == ' ') src++;

  /* arg3 */
  for(i=0;i<(len-1);i++)
  {
    if(*src == '>' || *src == '\0')
      break;
    *(arg3++) = *(src++);
  }
  *arg3 = '\0';
  return(3);
}

int get_userbypassword(name, address, password)
char *name;
char *address;
char *password;
{
  int i;
  for(i=0;i<MAX_USERS;i++)
  {
    if(strcmp(chat->user[i].password, password) == 0)
    {
      strncpy(name, chat->user[i].name, NAME_LEN);
      strncpy(address, chat->user[i].address, ADDR_LEN);
      return(i);
    }
  }
  return(-1);
}

#define MT_ALERT    0
#define MT_ERROR    1

int sendmessage(buf, maxlen, msg, type)
char *buf;
int maxlen;
char *msg;
int type;
{
  int i, len = 0, pos = 1;
  int msgs;
  msgs = (int)buf[0];
  for(i=0;i<msgs;i++)
  {
    int j, t;
    t = buf[pos++];
    j = strlen(&buf[pos]);
    pos += j + 1;
  }
  len = strlen(msg);
  if(len + pos < maxlen - 2)
  {
    buf[pos++] = (char)type;
    strcpy(&buf[pos], msg);
    buf[len + pos] = '\0';
    buf[0]++;
    return(1);
  }
  return(0);
}

void alertmessages()
{
  int i, j;
  if((j = (int)alertmsg[0]) > 0)
  {
    char *p, type;
    hasalert = 1;
    printf("<SCRIPT LANGUAGE=\"JavaScript\">\n");
    printf("<!--\n");
    printf("function showalert() {\n");
    p = &alertmsg[1];
    for(i=0;i<j;i++)
    {
      char amsg[MEMO_SIZE*2], *q = amsg;
      type = *p++;
      if(type == MT_ALERT) alertbtn = 1;
      while(*p != '\0')
      {
        if(sjis1st(*p))
        {
          *q++ = *p++;
          if(*p == '\\')
          {
            *q++ = '\\';
          }
          *q++ = *p++;
        }
        else
        {
          *q++ = *p++;
        }
      }
      *q = '\0';
      printf("alert(\"%s\")\n", amsg);
      p++;
    }
    printf("}\n");
    printf("//-->\n");
    printf("</SCRIPT>\n");
  }
}

void historyback()
{
    printf("<SCRIPT LANGUAGE=\"JavaScript\">\n");
    printf("<!--\n");
    printf("history.back()\n");
    printf("//-->\n");
    printf("</SCRIPT>\n");
}

/* 0: undenied 1:denied */
int denyhost(chat, dh, tim, name)
CHAT *chat;
IP dh;
int tim;
char *name;
{
  int i;
  for(i=0;i<MAX_DENYHOSTS;i++)
  {
    if(chat->denyhosts[i].address == dh)
    {
      chat->denyhosts[i].atime = 0;
      chat->denyhosts[i].address = 0;
      break;
    }
    if(chat->denyhosts[i].atime < now)
    {
      chat->denyhosts[i].atime = now + 3600*tim;
      chat->denyhosts[i].address = dh;
      strncpy(chat->denyhosts[i].name, name, NAME_LEN);
      break;
    }
  }
  chat->denyhosts[i].mid = userid;
  return(0);
}

#ifdef EXPORT_USER
void export_users(chat)
CHAT *chat;
{
  int i;

  /* truncate the export file */
  if(export_open(EXPORT_USER) == -1) return;
  export_close();

  for(i = 0; i < MAX_USERS; i++)
  {
    if(*chat->user[i].password != '\0' || chat->user[i].permit > 0)
    {
      if(export_user(EXPORT_USER, &chat->user[i]) == -1) break;
    }
  }
}
void import_users(chat)
CHAT *chat;
{
  int i = 0;
  if(import_open(EXPORT_USER) != -1)
  {
    while(i < MAX_USERS)
    {
      if(import_user(&chat->user[i]) == -1) break;
      if(*chat->user[i].name != '\0' &&
         (*chat->user[i].password != '\0' || chat->user[i].permit > 0) &&
         (chat->user[i].count > 3) )
      {
        chat->user[i].maxlines = DEF_MAX_LINES;
        i++;
      }
    }
    import_close();
  }
}
#endif /* EXPORT_USER */

#ifdef EXPORT_BBS
void export_bbss(chat)
CHAT *chat;
{
  int i;

  if(export_open(EXPORT_BBS) != -1)
  {
    for(i = 0; i < MAX_BBS && i < chat->bbs_line; i++)
    {
      int pline;
      if(chat->bbs_line >= MAX_BBS)
        pline = (chat->bbs_line + i)%MAX_BBS;
      else
        pline = i;
      if(*chat->bbs[pline].message != '\0')
      {
        if(export_text(chat->bbs[pline].name,  chat->bbs[pline].message,
                    chat->bbs[pline].atime) == 0) break;
      }
    }
    export_close();
  }
}

void import_bbss(chat)
CHAT *chat;
{
  int i;
  if(import_open(EXPORT_BBS) != -1)
  {
    chat->bbs_line = 0;
    for(i = 0; i < MAX_BBS; i++)
    {
      if(import_text(chat->bbs[i].name,  chat->bbs[i].message,
           &chat->bbs[i].atime) > 0)
        chat->bbs_line ++;
      else
        break;
    }
    import_close();
  }
}
#endif /* EXPORT_BBS */

#ifdef EXPORT_INI
void export_ini(chat)
CHAT *chat;
{
  FILE *fp;
  int i;
  fp = fopen(EXPORT_INI, "w");
  if(fp == NULL) return;
  fprintf(fp, "TITLE:%s\n", chat->title);
  fprintf(fp, "BGCOLOR:%s\n", chat->bgcolor);
  fprintf(fp, "BGIMAGE:%s\n", chat->bgimage);
  fprintf(fp, "TXCOLOR:%s\n", chat->txcolor);
  fprintf(fp, "LICOLOR:%s\n", chat->licolor);
  fprintf(fp, "VLCOLOR:%s\n", chat->vlcolor);
  fprintf(fp, "ALCOLOR:%s\n", chat->alcolor);
  fprintf(fp, "DNCOLOR:%s\n", chat->dncolor);
  fprintf(fp, "DTCOLOR:%s\n", chat->dtcolor);
  for(i = 0; i < MAX_LINKS; i++)
  {
    if(*chat->links[i].url != '\0')
      fprintf(fp, "LINK:%s,%s,%s\n", chat->links[i].url,
        chat->links[i].label_ja, chat->links[i].label_en);
  }
  fclose(fp);
}
void import_ini(chat)
CHAT *chat;
{
  FILE *fp;
  char buf[1024];
  int i, n = 0;
  fp = fopen(EXPORT_INI, "r");
  if(fp == NULL) return;
  while(fgets(buf, sizeof(buf), fp) != NULL)
  {
    for(i=strlen(buf)-1;i>=0;i--)
      if(buf[i] >= '\0' && buf[i] <= ' ') buf[i] = '\0';
      else break;
    if(strncmp(buf, "TITLE:", 6) == 0)
      strncpy(chat->title, &buf[6], TITLE_LEN);
    else if(strncmp(buf, "BGCOLOR:", 8) == 0)
      strncpy(chat->bgcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "BGIMAGE:", 8) == 0)
      strncpy(chat->bgimage, &buf[8], ADDR_LEN);
    else if(strncmp(buf, "TXCOLOR:", 8) == 0)
      strncpy(chat->txcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "LICOLOR:", 8) == 0)
      strncpy(chat->licolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "VLCOLOR:", 8) == 0)
      strncpy(chat->vlcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "ALCOLOR:", 8) == 0)
      strncpy(chat->alcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "DNCOLOR:", 8) == 0)
      strncpy(chat->dncolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "DTCOLOR:", 8) == 0)
      strncpy(chat->dtcolor, &buf[8], COLOR_LEN);
    else if(strncmp(buf, "LINK:", 5) == 0)
    {
      if(n < MAX_LINKS)
      {
        char *jp, *en;
        if((jp = (char *)strchr(buf, ',')) == NULL) continue;
        *jp = '\0';
        jp++;
        if((en = (char *)strchr(jp, ',')) != NULL)
        {
          *en = '\0';
          en++;
        }
        else
        {
          en = jp;
        }
        strncpy(chat->links[n].url, &buf[5], ADDR_LEN);
        strncpy(chat->links[n].label_ja, jp, LABEL_LEN);
        strncpy(chat->links[n].label_en, en, LABEL_LEN);
        n++;
      }
    }
  }
  fclose(fp);
}
#endif /* EXPORT_INI */

#ifdef EXPORT_CHAT
void export_chat(chat, line, data)
CHAT *chat;
int line;
TEXT *data;
{
  FILE *fp;
  fp = fopen(EXPORT_CHAT, "r+");
  if(fp != NULL)
  {
    fwrite((void *)&line, sizeof(int), 1, fp);
    fseek(fp, sizeof(int)+sizeof(TEXT)*line, SEEK_SET);
    fwrite(data, sizeof(TEXT), 1, fp);
    fclose(fp);
  }
}
void import_chat(chat)
CHAT *chat;
{
  FILE *fp;
  int i;
  fp = fopen(EXPORT_CHAT, "r");
  if(fp != NULL)
  {
    fread((void *)&chat->text_line, sizeof(int), 1, fp);
    for(i = 0; i < MAX_LINES; i++)
    {
      if(fread(&chat->text[i], sizeof(TEXT), 1, fp) < 1) break;
    }
    fclose(fp);
    chat->text_line = (++chat->text_line)%MAX_LINES;
  }
}
#endif /* EXPORT_CHAT */


int specialcmd(chat, message)
CHAT *chat;
char *message;
{
  int i, j;
  char arg1[1024], arg2[1024], arg3[1024];
  int argc;

  if(*message != '<' && *message != '\\') return(0);

  if(*message == '\\')
  {
    strcpy(arg1, "\\");
    argc = 1;
  }
  else
    argc = get_tagparam(message, 1024, arg1, arg2, arg3);

  scstrlwr(arg1);

  if(strcmp(arg1, "who") == 0)
  {
    userproperty = 1;
    return(1);
  }
  else if(strcmp(arg1, "cancel") == 0)
  {
    int n = 1;
    if(argc == 1)
    {
      strncpy(arg2, name, sizeof(arg2));
      if(userid > 0 && chat->user[userid].count > 0)
        chat->user[userid].count --;
    }
    else if(permission < 2) goto cmd_error;

    if(argc == 3)
      n = atoi(arg3);

    for(i = 1; i < MAX_LINES; i++)
    {
      j = (MAX_LINES+chat->text_line-i)%MAX_LINES;
      if((strcmp(chat->text[j].name, arg2) == 0) &&
         (*chat->text[j].message != '\0') )
      {
        *chat->text[j].message = '\0';
        if(--n > 0) continue;
        break;
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "log") == 0)
  {
    int log;
    if(permission < 1 || argc < 2) goto cmd_error;
    log = atoi(arg2);
    if(log < 1 || log > MAX_LINES) goto cmd_error;
    chat->user[userid].maxlines = log;
    maxlines = log;
    return(1);
  }
  else if(strcmp(arg1, "color") == 0)
  {
    if(argc < 2)
      *chat->user[userid].namecolor = '\0';
    else
    {
      if(isvalidcolor(arg2, def_bgcolor) == 0) goto cmd_error;
      if(argc > 2)
        if(isvalidcolor(arg3, def_bgcolor) == 0) goto cmd_error;
      strncpy(chat->user[userid].namecolor, arg2, COLOR_LEN);
    }
    if(argc == 3)
    {
      strncpy(chat->user[userid].textcolor, arg3, COLOR_LEN);
    }
    return(1);
  }
  else if(strcmp(arg1, "deny") == 0)
  {
    IP dh;
    int tim = REFUSETIME;
    int ds = -1;
    if(permission < 2 || argc < 2) goto cmd_error;
    dh = str2IP(arg2);
    if(argc == 3)
      tim = atoi(arg3);
    if(dh != -1)
    {
      ds = denyhost(chat, dh, tim, "");
    }
    else
    {
      for(i=0;i<MAX_USERS;i++)
      {
        if(strcmp(chat->user[i].name, arg2) == 0)
        {
          ds = denyhost(chat, chat->user[i].haddressip, tim, arg2);
          chat->user[i].login = 2;
        }
      }
    }
    if(ds == -1)
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_COMMANDFAIL), MT_ERROR);
      return(1);
    }
    return(1);
  }
  else if(strcmp(arg1, "reset") == 0)
  {
    if(argc == 1) /* clear password */
    {
        *chat->user[userid].password = '\0';
        chat->user[userid].mid = userid;
        if(chat->user[userid].permit == 2)
          chat->user[userid].permit = 1;
        export_user(EXPORT_USER, &chat->user[userid]);
        userreseted = 1;
        return(1);
    }
    if(permission < 2 || argc < 2) goto cmd_error;
    for(i = 0; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0 &&
         chat->user[i].permit >= 0)
      {
        chat->user[i].permit = 0;
        *chat->user[i].password = '\0';
        chat->user[i].mid = userid;
        export_user(EXPORT_USER, &chat->user[i]);
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "permit") == 0)
  {
    if(permission < 1 || argc < 2) goto cmd_error;
    for(i = 0; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0)
      {
        if(*chat->user[i].password != '\0')
        {
          if(chat->user[i].permit < permission)
          {
            chat->user[i].permit ++;
            chat->user[i].mid = userid;
            export_user(EXPORT_USER, &chat->user[i]);
          }
          break;
        }
        else /* non passworded user */
        {
          if((chat->user[i].login == 1) &&
             (chat->user[i].atime > now - AUTOLOGOUT))
          {
            chat->user[i].permit = 1;
            chat->user[i].mid = userid;
          }
        }
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "topic") == 0)
  {
    char *p = &message[6];
    int i;
    if(permission < 1) goto cmd_error;
    while(*p == ' ') p++;
    if(*p == '\\')
      roma2kana(arg2, ++p, sizeof(arg2));
    else
      strncpy(arg2, p, sizeof(arg2));
    if((i = strlen(arg2)) > 0)
      if(arg2[i - 1] == '>') arg2[i - 1] = '\0';
    strncpy(chat->topic.name, name, NAME_LEN);
    strncpy(chat->topic.message, arg2, MSG_SIZE);
    chat->topic.atime = now;
    return(1);
  }
  else if(strcmp(arg1,"alert") == 0)
  {
    char msg[MEMO_SIZE];
    int i, j=0;
    if(argc < 3) goto cmd_error;
    if(strcmp(arg2, MANAGER) == 0)
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_CANNOTSEND), MT_ERROR);
      return(1);
    }
    for(i=0;arg3[i] != '\0';i++)
      if(arg3[i] == '"') arg3[i] = ' ';
    printdate(msg, now);
    strcat(msg, "\\n");
    strcat(msg, name);
    strcat(msg, "> ");
    i = strlen(msg);
    strncpy(&msg[i], arg3, sizeof(msg) - i);
    for(i = 0; i < MAX_USERS; i++)
    {
      if(strcmp(chat->user[i].name, arg2) == 0 ||
         (ismanager() && *arg2 == '*'))
      {
        if(!sendmessage(chat->user[i].memo, MEMO_SIZE, msg, MT_ALERT))
        {
          sendmessage(alertmsg, sizeof(alertmsg), MSG(M_CANNOTSEND), MT_ERROR);
          break;
        }
        j ++;
      }
    }
    if(j == 0)
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_CANNOTSEND), MT_ERROR);
    }
    return(1);
  }
  else if(strcmp(arg1, "su") == 0)
  {
    if(*password != '\0' &&
       strcmp(arg2, MANAGER "/" PASSWORD) == 0 && userid >= 0)
    {
      chat->user[userid].permit = 2;
      chat->user[userid].mid = userid;
      export_user(EXPORT_USER, &chat->user[userid]);
      return(1);
    }
    goto cmd_error;
  }
  else if(strcmp(arg1, "export") == 0)
  {
    if(permission < 2) goto cmd_error;
#ifdef EXPORT_USER
    export_users(chat);
#endif
#ifdef EXPORT_BBS
    export_bbss(chat);
#endif
  }
  else if(strcmp(arg1, "r") == 0 || strcmp(arg1, "\\") == 0)
  {
    char *p, kana[MSG_SIZE] = "";
    if(*arg1 == '\\')
      roma2kana(kana, &message[1], sizeof(kana));
    else if((p = (char *)strchr(message, '>')) != NULL)
      roma2kana(kana, ++p, sizeof(kana));
    strncpy(message, kana, MSG_SIZE);
  }
  return(0);
  
cmd_error:
  sendmessage(alertmsg, sizeof(alertmsg), MSG(M_COMMANDFAIL), MT_ERROR);
  return(1);
}

int specialbbscmd(chat, message)
CHAT *chat;
char *message;
{
  char arg1[1024], arg2[1024], arg3[1024];
  int argc;

  if(*message != '<') return(0);

  argc = get_tagparam(message, 1024, arg1, arg2, arg3);
  if(strcmp(arg1, "cancel") == 0)
  {
    int i, j;
    for(i=1;i<=MAX_BBS;i++)
    {
      j = (MAX_BBS+chat->bbs_line-i)%MAX_BBS;
      if((strcmp(chat->bbs[j].name, name) == 0) &&
         (*chat->bbs[j].message != '\0') )
      {
        strncpy(message, chat->bbs[j].message, MSG_SIZE);
        *chat->bbs[j].message = '\0';
        chat->bbs[j].atime = 0;
        break;
      }
    }
    return(1);
  }
  else if(strcmp(arg1, "delete") == 0)
  {
    int i;
    i = atoi(arg2);
    strncpy(message, chat->bbs[i].message, MSG_SIZE);
    message[strlen(chat->bbs[i].message)] = '\0';
    *chat->bbs[i].message = '\0';
    chat->bbs[i].atime = 0;
    return(1);
  }
  else if(strcmp(arg1, "r") == 0 || strcmp(arg1, "\\") == 0 )
  {
    char *p, kana[MSG_SIZE] = "";
    if(*arg1 == '\\')
      roma2kana(kana, &message[1], sizeof(kana));
    if((p = (char *)strchr(message, '>')) != NULL)
      roma2kana(kana, ++p, sizeof(kana));
    strncpy(message, kana, MSG_SIZE);
  }
  return(0);
}

int shmid;
int semid;
static struct sembuf semops[1];
#ifndef SEM_KEY
#define SEM_KEY SHM_KEY
#endif

CHAT *get_chat()
{
  CHAT *chat;
  if((semid = semget(SEM_KEY, 1, 0666)) == EOF)
  {
    if((semid = semget(SEM_KEY, 1, 0666|IPC_CREAT)) == EOF)
    {
      scerr(MSG(M_CANNOTUSE));
    }
    semctl(semid, 0, SETVAL, 1);
  }
  if((shmid=shmget(SHM_KEY, sizeof(CHAT), 0666)) == EOF)
  {
    int retry = 0;
retry_get:
    if((shmid=shmget(SHM_KEY, sizeof(CHAT), 0666|IPC_CREAT)) == EOF)
    {
      if((shmid=shmget(SHM_KEY, 0, 0666)) == EOF)
      {
        scerr(MSG(M_CANNOTUSE));
      }
      if(shmctl(shmid, IPC_RMID, 0) == EOF)
      {
        scerr(MSG(M_CANNOTUSE));
      }
      if(retry == 0)
      {
        retry++;
        goto retry_get;
      }
      scerr(MSG(M_CANNOTUSE));
    }
    if((int)(chat = (CHAT *)shmat(shmid, 0, 0)) == EOF)
    {
      scerr(MSG(M_CANNOTUSE));
    }
    init_memory(chat); /* initialize work area */
    shmdt(chat);
    chat = NULL;
  }
  if((int)(chat = (CHAT *)shmat(shmid, 0, 0)) == EOF)
  {
    scerr(MSG(M_CANNOTUSE));
  }
  if(chat->version != memversion)
    reset_chat(chat);
  return(chat);
}

void release_chat(chat)
CHAT *chat;
{
  shmdt(chat);
  chat = NULL;
}

void lock_chat(chat)
CHAT *chat;
{ 
  signal(SIGALRM, sigalm);
  alarm(TIMEOUT);
  semops[0].sem_num = 0;
  semops[0].sem_op = -1; /* Lock */
  semops[0].sem_flg = 0;
  semop(semid, semops, 1);
}

void unlock_chat(chat)
CHAT *chat;
{
  signal(SIGALRM, SIG_IGN);
  semops[0].sem_num = 0;
  semops[0].sem_op = 1; /* Unlock */
  semops[0].sem_flg = 0;
  semop(semid, semops, 1);
}

void reset_chat(chat)
CHAT *chat;
{
  if((int)shmctl(shmid, IPC_RMID, 0) == EOF)
  {
    scerr(MSG(M_CANNOTUSE));
  }
  if((int)semctl(semid, 0, IPC_RMID, 0) == EOF)
  {
    scerr(MSG(M_CANNOTUSE));
  }
#ifdef EXPORT_CHAT
  {
    FILE *fp;
    if((fp = fopen(EXPORT_CHAT, "w")) != NULL) fclose(fp);
  }
#endif /* EXPORT_CHAT */
  
  scerr(MSG(M_INITIALIZED));
}


CHATMODE get_mode(mode)
char *mode;
{
  if(strcmp(mode, "entry") == 0)
    return(m_entry);
  else if(strcmp(mode, "input") == 0)
    return(m_input);
  else if(strcmp(mode, "chattext") == 0)
    return(m_chattext);
  else if(strcmp(mode, "bbs") == 0)
    return(m_bbs);
  else if(strcmp(mode, "logout") == 0)
    return(m_logout);
  else if(strcmp(mode, "customize") == 0)
    return(m_customize);
  else if(strcmp(mode, "reset") == 0)
    return(m_reset);
  else if(strcmp(mode, "debug") == 0)
    return(m_debug);
  else if(strcmp(mode, "doc") == 0)
    return(m_doc);
  return(m_init);
}

#define IFSETSPARAM(x,y) \
  if(strncmp(text,y "=",sizeof(y))==0 && \
     strlen(text)>sizeof(y)) \
      strncpy((x), &text[sizeof(y)], sizeof(x))

#define IFCATSPARAM(x,y) \
  if(strncmp(text,y "=",sizeof(y))==0 && \
     strlen(text)>sizeof(y)) \
      strncat((x), &text[sizeof(y)], sizeof(x) - strlen(x))

#define IFSETIPARAM(x,y) \
  if(strncmp(text,y "=",sizeof(y))==0 && \
     strlen(text)>sizeof(y)) \
      x = atoi(&text[sizeof(y)])

void main(argc, argv)
int argc;
char **argv;
{
  char *param;
  char text[MSG_SIZE];
  char *src, *proxy;
  int i;
  int chatlines = 0;
  time_t logintime;
  IP httpfor = 0;

  signal(SIGTERM, sigterm);
  signal(SIGINT, sigterm);
  signal(SIGSEGV, sigterm);
  signal(SIGPIPE, sigterm);

#ifdef COOKIE
  strncpy(cookie, get_env("HTTP_COOKIE"), sizeof(cookie));
#endif /* COOKIE */
  strncpy(agent, get_env("HTTP_USER_AGENT"), sizeof(agent));
  strncpy(haddress, get_env("REMOTE_HOST"), sizeof(haddress));
  if(*(proxy = get_env("HTTP_X_FORWARDED_FOR")) != '\0') /* via Squid */
  {
    if(isIP(proxy)) httpfor = str2IP(proxy);
  }
  else if(*(proxy = get_env("HTTP_FROM")) != '\0') /* via DeleGate4.xx */
  {
    proxy += 2;
    if(isIP(proxy)) httpfor = str2IP(proxy);
  }
  if(httpfor != 0)
  {
    int len = strlen(haddress), maxlen = sizeof(haddress);
    /* private nets (RFC 1918) */
    if(!(compareIP(httpfor, str2IP("192.168.0.0"), 0xffff0000) == 0 ||
         compareIP(httpfor, str2IP("172.16.0.0"), 0xfff00000) == 0 ||
         compareIP(httpfor, str2IP("10.0.0.0"), 0xff000000) == 0))
    {
      if(len < maxlen - 3)
      {
        strcpy(&haddress[len], "->");
        strncpy(&haddress[len+2], proxy, maxlen - len - 2);
      }
    }
  }
    
  haddressip = str2IP(get_env("REMOTE_ADDR"));
  if(strstr(agent, "Mozilla") != NULL || strstr(agent, "Plala") != NULL)
  {
    sjis = 1;
  }
  if(strstr(agent, "; U)") != NULL)
  {
    strcpy(lang, "en");
  }
  now = time(NULL);
  nowtm = localtime(&now);
  chat_mode = m_init;

  chat = get_chat();
  if(chat == NULL)
  {
    scerr(MSG(M_CANNOTUSE));
  }
  strncpy(def_title, chat->title, sizeof(def_title));
  strncpy(def_bgcolor, chat->bgcolor, sizeof(def_bgcolor));
  strncpy(def_txcolor, chat->txcolor, sizeof(def_txcolor));
  strncpy(def_licolor, chat->licolor, sizeof(def_licolor));
  strncpy(def_vlcolor, chat->vlcolor, sizeof(def_vlcolor));
  strncpy(def_alcolor, chat->alcolor, sizeof(def_alcolor));
  strncpy(def_dncolor, chat->dncolor, sizeof(def_dncolor));
  strncpy(def_dtcolor, chat->dtcolor, sizeof(def_dtcolor));
  SETBODY(body, def_bgcolor, def_bgimage, def_txcolor, def_licolor, def_vlcolor, def_alcolor);
  if(*def_title == '\0') title = MSG(M_TITLE);
  else                   title = def_title;

  if((param = getcgiarg()) == NULL)
    exit(1);

  src = param;
  while(*src != '\0')
  {
    src = getparam(text, sizeof(text), src);
    IFSETSPARAM(message, "chat");
    else IFSETIPARAM(refresh, "refresh");
    else IFSETIPARAM(maxlines, "maxlines");
    else IFSETSPARAM(bgcolor, "bgcolor");
    else IFSETSPARAM(txcolor, "txcolor");
    else IFSETSPARAM(cpassword, "password");
    else IFSETSPARAM(address, "address");
    else IFSETSPARAM(referer, "referer");
    else IFSETSPARAM(def_title, "def_title");
    else IFSETSPARAM(def_bgcolor, "def_bgcolor");
    else IFSETSPARAM(def_txcolor, "def_txcolor");
    else IFSETSPARAM(def_bgimage,  "def_bgimage");
    else IFSETSPARAM(def_licolor, "def_licolor");
    else IFSETSPARAM(def_vlcolor, "def_vlcolor");
    else IFSETSPARAM(def_alcolor, "def_alcolor");
    else IFSETSPARAM(def_dncolor, "def_dncolor");
    else IFSETSPARAM(def_dtcolor, "def_dtcolor");
    else IFSETSPARAM(mode, "mode");
    else IFSETSPARAM(lang, "lang");
    else IFSETSPARAM(frame, "frame");
    else IFSETSPARAM(secure, "secure");
    else IFSETIPARAM(room, "room");
    else if(strncmp(text, "link_url=", 9) == 0)
    {
      if(link_items < MAX_LINKS)
        strncpy(link_url[link_items], &text[9], ADDR_LEN);
    }
    else if(strncmp(text, "link_label=", 11) == 0)
    {
      if(link_items < MAX_LINKS)
      {
        strncpy(link_label[link_items], &text[11], LABEL_LEN);
        link_items++;
      }
    }
    else if(strncmp(text, "name=", 5) == 0 && strlen(text) > 5)
    {
      char *n = &text[5], *p = NULL;
      while((*n == ' ') && (*n != '\0')) n++;
      strncpy(name, n, sizeof(name) - 1);
      for(i=strlen(name)-1;i>0;i--)
        if(name[i]>=0&&name[i]<=0x20) name[i] = '\0';
      if((p = (char *)strchr(name, '/')) != NULL) /* with a password */
      {
        strncpy(password, p+1, sizeof(password));
        code(cpassword, password);
        *p = '\0';
        if(*name != '\0') *address = '\0';
      }
      killtag(name);
      if(*name == '\0' && *password != '\0') /* password only */
      {
        if(get_userbypassword(name, address, password) == -1 ||
           strcmp(password, PASSWORD) == 0)
        {
          password[0] = '\0';
          address[0] = '\0';
          name[0] = '\0';
          sendmessage(alertmsg, sizeof(alertmsg), MSG(M_INVALIDPASS), MT_ERROR);
          back = 1;
        }
      }
    }
    else if(strncmp(text, "autoclear", 9) == 0 && strlen(text) >= 9)
    {
      autoclear = 1;
    }
    else if(strncmp(text, "auto", 4) == 0 && strlen(text) >= 4)
    {
      autoreload = 1;
    }
  }
  decode(password, cpassword);
  killtag(address);
  chat_mode = get_mode(mode);
  if(*bgcolor != '\0')
    body.bgcolor = bgcolor;
  if(*txcolor != '\0')
    body.txcolor = txcolor;
  if(*lang != '\0')
  {
    if(strcmp(lang, "ja") == 0)
    {
      LANG = MSG_JA;
      CONV = 0;
    }
    else if(strcmp(lang, "en") == 0)
    {
      LANG = MSG_EN;
      CONV = 1;
    }
    else if(strcmp(lang, "enj") == 0)
    {
      LANG = MSG_EN;
      CONV = 0;
    }
  }

  /* check deny hosts */
  if(*name != '\0' && !ismanager() && isdenyhost(chat, haddressip))
  {
    lock_chat(chat);
    userid = setuserinfo();
    if(userid >= 0)
    {
      if(chat->user[userid].permit < 2)
        chat->user[userid].login = 2;
    }
    unlock_chat(chat);
    if(chat->user[userid].permit < 2)
      scerr(MSG(M_CANNOTUSE));
  }

  if(*name == '\0') /* entry form */
  {
    chat_mode = m_init;
  }

  switch(chat_mode)
  {
  case m_entry:
    lock_chat(chat);
    userid = setuserinfo();
    unlock_chat(chat);
    release_chat(chat);
    enter_base();
    break;
  case m_input:
    input_form(chat);
    break;
  case m_chattext:
  case m_bbs:
  case m_logout:
    break;
  case m_reset:
    if(!ismanager())
      scerr(MSG(M_CANNOTUSE));
    reset_chat(chat);
    break;
  case m_debug:
    if(!ismanager())
      scerr(MSG(M_CANNOTUSE));
    printdebug(chat);
    chat_exit(0);
    break;
  case m_customize:
    if(ismanager())
    {
      lock_chat(chat);
      if(link_items == 0)
      {
        strncpy(chat->title, def_title, sizeof(def_title));
        strncpy(chat->bgcolor, def_bgcolor, sizeof(def_bgcolor));
        strncpy(chat->txcolor, def_txcolor, sizeof(def_txcolor));
        strncpy(chat->licolor, def_licolor, sizeof(def_licolor));
        strncpy(chat->vlcolor, def_vlcolor, sizeof(def_vlcolor));
        strncpy(chat->alcolor, def_alcolor, sizeof(def_alcolor));
        strncpy(chat->dncolor, def_dncolor, sizeof(def_dncolor));
        strncpy(chat->dtcolor, def_dtcolor, sizeof(def_dtcolor));
      }
      else
      {
        int i;
        for(i = 0; i < link_items; i++)
        {
          strncpy(chat->links[i].url, link_url[i], ADDR_LEN);
          strncpy(chat->links[i].label_ja, link_label[i], LABEL_LEN);
        }
      }
#ifdef EXPORT_INI
      export_ini(chat);
#endif
      unlock_chat(chat);
    }
    release_chat(chat);
    enter_base();
    break;
  case m_init:
  default:
    entry_form(chat);
    break;
  }

  local = (CHAT *)malloc(sizeof(CHAT));

  lock_chat(chat);

  userid = setuserinfo();
  permission = chat->user[userid].permit;

  if(chat_mode == m_logout)
  {
    unlock_chat(chat);
    sleep(LOGOUTWAIT);
    lock_chat(chat);
    logintime = chat->user[userid].logintime;
  }

  if(chat_mode == m_bbs)
  {
    if(*message != '\0' && permission > 0)
    {
      if(specialbbscmd(chat, message) == 0)
      {
        int bbs_line = chat->bbs_line%MAX_BBS;
        *chat->bbs[bbs_line].message = '\0';
        chat->bbs[bbs_line].atime = now;
        strncpy(chat->bbs[bbs_line].name, name, NAME_LEN);
        strncpy(chat->bbs[bbs_line].namecolor, namecolor, COLOR_LEN);
        strncpy(chat->bbs[bbs_line].textcolor, textcolor, COLOR_LEN);
        strncpy(chat->bbs[bbs_line].message, message, MSG_SIZE);
        if(strcmp(secure, "on") == 0)
          chat->bbs[bbs_line].flags |= SECURE_BBS;
        else
          chat->bbs[bbs_line].flags &= ~SECURE_BBS;
        chat->bbs_line++;
        *message = '\0';
      }
      export_bbss(chat);
    }
    chat->user[userid].btime = now;
    if(*chat->user[userid].password != '\0')
      export_user(EXPORT_USER, &chat->user[userid]);
    memcpy(local, chat, sizeof(CHAT));

    unlock_chat(chat);
    release_chat(chat);

    printbbs(local, MAX_BBS);
    chat_exit(0);
  }
  if(chat_mode == m_chattext)
  {
    if(*message != '\0')
    {
      if(specialcmd(chat, message) == 0)
      {
        *chat->text[chat->text_line].message = '\0';
        chat->text[chat->text_line].atime = now;
        strncpy(chat->text[chat->text_line].name, name, NAME_LEN);
        strncpy(chat->text[chat->text_line].namecolor, namecolor, COLOR_LEN);
        strncpy(chat->text[chat->text_line].textcolor, textcolor, COLOR_LEN);
        strncpy(chat->text[chat->text_line].message, message, MSG_SIZE);
        if(CONV == 1)
          chat->text[chat->text_line].flags |= TEXT_ITALIC;
        else
          chat->text[chat->text_line].flags &= ~TEXT_ITALIC;
        chat->text[chat->text_line].room = room;
#ifdef EXPORT_CHAT
        export_chat(chat, chat->text_line, &chat->text[chat->text_line]);
#endif /* EXPORT_CHAT */
        chat->text_line = (++chat->text_line)%MAX_LINES;
        chat->user[userid].count ++;
      }
    }
  }
  chatlines = NEWUSER_LINES + chat->user[userid].count;
  /* get alert messages */
  if((int)chat->user[userid].memo[0] > 0 && alertmsg[0] == '\0')
  {
    memcpy(alertmsg, chat->user[userid].memo, MEMO_SIZE);
    chat->user[userid].memo[0] = '\0';
    for(i=0;i<MAX_USERS;i++)
    { /* clear other same name's alert */
      if((strcmp(chat->user[i].name, name) == 0) && 
         ((int)chat->user[i].memo[0] > 0))
      {
        chat->user[i].memo[0] = '\0';
      }
    }
  }
  memcpy(local, chat, sizeof(CHAT));

  unlock_chat(chat);
  release_chat(chat);

  if(chat_mode == m_logout)
  {
    header(local, 0);
    printf("<HR>\n");
    printf("<H3>%s</H3>\n", MSG(M_GOODBYE));
    printf("%s: ", MSG(M_LOGINTIME));
    printdate(NULL, logintime);
    printf("<BR>\n");
    printf("%s: ", MSG(M_LOGOUTTIME));
    printdate(NULL, now);
  }
  else
  {
    header(local, 1);
  }
  if(maxlines == 0) maxlines = DEF_MAX_LINES;
  if(chatlines > MAX_LINES) chatlines = MAX_LINES;
  if(chatlines > maxlines) chatlines = maxlines;
  printchat(local, chatlines);

  footer();
  free(local);
  exit(0);
}

void print_content()
{
  printf("Content-type: text/html\n");
#ifdef COOKIE
  if(*cookie == '\0' && *name != '\0')
    printf("Set-Cookie: name=%s; expires=31-12-99 23:59:59 GMT\n", name);
#endif /* COOKIE */
  printf("\n");
  printf("<HTML>\n");
  printf("<HEAD>\n");
  if(LANG == MSG_JA && chat_mode != m_bbs)
    printf("<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html;CHARSET=x-sjis\">\n");
}

void print_common(mode)
char *mode;
{
  char pname[256];
  char paddr[256];
  if(sjis)
  {
    strncpy(pname, name, sizeof(pname));
    strncpy(paddr, address, sizeof(paddr));
  }
  else
  {
    sjis2euc(pname, name);
    sjis2euc(paddr, address);
  }

  hidden_out("mode", mode);
  hidden_out("lang", lang);
  hidden_out("name", pname);
  hidden_out("address", paddr);
  hidden_out("password", cpassword);
  hidden_out("bgcolor", bgcolor);
  hidden_out("txcolor", txcolor);
}

void entry_form(chat)
CHAT *chat;
{
  header(chat, 0);
  if(autoreload)
  {
    printf("</BODY></HTML>\n");
    chat_exit(0);
  }
  printtopic(chat);
  printf("<HR>\n");
  printf("<H3>%s</H3>\n", MSG(M_USAGE));
  printf("<FORM METHOD=\"%s\" ACTION=\"%s\">\n",
    cgimtd, cgibin);
  printf("<TABLE>\n");
  printf("<TR><TD>%s</TD>\n", MSG(M_NAME));
  printf("<TD><INPUT TYPE=\"TEXT\" NAME=\"name\" SIZE=\"15\"");
    printf(" MAXLENGTH=\"40\"><BR></TD></TR>\n");
  printf("<TR><TD>%s</TD>", MSG(M_ADDRESS));
  printf("<TD><INPUT TYPE=\"TEXT\" NAME=\"address\" SIZE=\"40\"");
    printf(" MAXLENGTH=\"80\">\n");
  printf("<BR></TD></TR>\n");
  printf("</TABLE>\n");
  printf(" %s:<SELECT NAME=\"refresh\">", MSG(M_RELOAD));
  printf("<OPTION>5<OPTION>10<OPTION>15<OPTION SELECTED>20");
  printf("<OPTION>30<OPTION>50</SELECT>\n");
  printf("<INPUT TYPE=\"CHECKBOX\" NAME=\"autoclear\"");
  if(autoclear) printf(" CHECKED");
  printf(">%s", MSG(M_AUTOCLEAR));

  hidden_out("mode", "entry");
  hidden_out("lang", lang);
  hidden_out("bgcolor", bgcolor);
  hidden_out("txcolor", txcolor);
  if(room != 0)
    hidden_out("room", "1");

  if(getenv("HTTP_REFERER") != NULL)
  {
    char tmp[ADDR_LEN], *p;
    strncpy(tmp, getenv("HTTP_REFERER"), ADDR_LEN);
    if((p = (char *)strchr(tmp, '?')) != NULL) *p = '\0';
    hidden_out("referer", tmp);
  }
  printf("<BR>\n");
  printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_JOIN));
  printf("<INPUT TYPE=\"RESET\" VALUE=\"%s\">\n", MSG(M_CLEAR));
  printf("</FORM>\n");
  printf("<HR>\n");
  printf("%s", schatinfo[LANG]);
  printlinks(chat);
  footer();
  chat_exit(0);
}

void noframe_input()
{
  printf("<FORM METHOD=\"%s\" ACTION=\"%s\">\n",
    cgimtd, cgibin);
  hidden_out("frame", "no");
  print_common("chattext");
  printf("<INPUT TYPE=\"TEXT\" NAME=\"chat\" SIZE=\"50\"");
    printf(" MAXLENGTH=\"1023\">\n");
  printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_MESSAGE));
  printf("<INPUT TYPE=\"RESET\" VALUE=\"%s\">\n", MSG(M_CLEAR));
  printf("</FORM>\n");
}

void enter_base()
{
  char pname[256];
  char paddr[256];
  char ename[256];
  char eaddr[256];
  char pbgcolor[41];
  char ptxcolor[41];
  int inputsize = 20;
  if(sjis)
  {
    strncpy(ename, name, sizeof(ename));
    strncpy(eaddr, address, sizeof(eaddr));
  }
  else
  {
    sjis2euc(ename, name);
    sjis2euc(eaddr, address);
  }
  paramencode(pname, ename);
  paramencode(paddr, eaddr);
  paramencode(pbgcolor, bgcolor);
  paramencode(ptxcolor, txcolor);
  if(ismanager()) inputsize = 50;
  print_content();
  printf("<TITLE>"); nohtmlout(title);
  printf("("); htmlout(name, 1); printf(")</TITLE>\n");
  printf("</HEAD>\n");
  printf("<FRAMESET ROWS=\"%d%%,*\">\n", inputsize);

  printf("<FRAME SRC=\"%s?", cgibin);
  printf("mode=input&name=%s&address=%s", pname, paddr);
  if(room == 1) printf("&room=1");
  printf("&password=%s", cpassword);
  if(*bgcolor != '\0')
    printf("&bgcolor=%s", pbgcolor);
  if(*txcolor != '\0')
    printf("&txcolor=%s", ptxcolor);
  if(*lang != '\0')
    printf("&lang=%s", lang);
  if(autoclear)
    printf("&autoclear=on");
  printf("\" NAME=\"chatin\">\n");

  printf("<FRAME SRC=\"%s?", cgibin);
  printf("mode=chattext&name=%s&address=%s", pname, paddr);
  if(room == 1) printf("&room=1");
  printf("&password=%s", cpassword);
  if(*bgcolor != '\0')
    printf("&bgcolor=%s", pbgcolor);
  if(*txcolor != '\0')
    printf("&txcolor=%s", ptxcolor);
  if(*lang != '\0')
    printf("&lang=%s", lang);
  printf("\" NAME=\"chattext\">\n");

  printf("</FRAMESET>\n");
  printf("<NOFRAMES>\n");
  body_out(&body);
  printf("<a href=\"%s?mode=chattext&frame=no&name=%s&address=%s&password=%s&lang=%s\">",
    cgibin, pname, paddr, cpassword, lang);
  htmlout(title, 1);
  printf("</a>\n");
  printf("</BODY>\n");
  printf("</NOFRAMES>\n");
  printf("</HTML>\n");
  chat_exit(0);
}

void input_form(chat)
CHAT *chat;
{
  int i;

  print_content();
  printf("</HEAD>\n");

  if(autoclear)
  {
    printf("<SCRIPT LANGUAGE=\"JavaScript\">\n");
    printf("function erase() {\n");
    printf("  document.forms[0].chat.value = document.forms[0].chat2.value;\n");
    printf("  document.forms[0].chat2.value = '';\n");
    printf("}\n");
    printf("function undo() {\n");
    printf("  document.forms[0].elements[1].value = document.forms[0].elements[0].value;\n");
    printf("}\n");
    printf("</SCRIPT>\n");
  }


  body_out(&body);
  if(autoclear)
  {
    printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"chattext\" ONSUBMIT=\"erase()\">\n",
      cgimtd, cgibin);
    printf("<INPUT TYPE=\"HIDDEN\" NAME=\"chat\" value=\"\">");
    printf("<INPUT TYPE=\"TEXT\" NAME=\"chat2\" SIZE=\"50\"");
      printf(" MAXLENGTH=\"1023\">\n");
  }
  else
  {
    printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"chattext\">\n",
      cgimtd, cgibin);
    printf("<INPUT TYPE=\"TEXT\" NAME=\"chat\" SIZE=\"50\"");
      printf(" MAXLENGTH=\"1023\">\n");
  }
  if(room == 1)
    printf("<INPUT TYPE=\"HIDDEN\" NAME=\"room\" VALUE=\"%d\">\n", room);
  print_common("chattext");
  printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_MESSAGE));
  printf("<INPUT TYPE=\"RESET\" NAME=\"reset\" VALUE=\"%s\">\n", MSG(M_CLEAR));
  if(autoclear)
    printf("<INPUT TYPE=\"BUTTON\" VALUE=\"%s\" ONCLICK=\"undo()\">\n", MSG(M_REEDIT));
  printf("</FORM><HR>\n");

  if(*password != '\0')
  {
    printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"chattext\">\n",
      cgimtd, cgibin);
    printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"who\">\n");
    printf("<INPUT TYPE=\"HIDDEN\" NAME=\"chat\" value=\"<who>\">");
    print_common("chattext");
    printf("</FORM>\n");
  }

  if(strstr(agent, "MSIE") != NULL)
  {
    printf("<FORM METHOD=\"GET\" ACTION=\"%s\" TARGET=\"_top\">\n", cgibin);
  }
  else
  {
    printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"_top\">\n",
      cgimtd, cgibin);
  }
  printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_LEAVE));
  print_common("logout");
  printf("</FORM>\n");

  printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"_blank\">\n",
    cgimtd, cgibin);
  printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_BBS));
  print_common("bbs");
  printf("</FORM>\n");

  if(!ismanager())
  {
    printlinks(chat);
  }
  else
  { /* special costomizeable form */
    printf("<HR>\n");
    printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"_top\">\n",
      cgimtd, cgibin);
    printf("%s:<INPUT TYPE=\"TEXT\" NAME=\"def_title\"", MSG(M_MODTITLE));
      printf(" VALUE=\"");
      print_string(def_title);
      printf("\" SIZE=20><BR>\n");
    printf("%s:<INPUT TYPE=\"TEXT\" NAME=\"def_bgcolor\"", MSG(M_MODBGCOLOR));
      printf(" VALUE=\"%s\" SIZE=10>\n", def_bgcolor);
    printf("%s:<INPUT TYPE=\"TEXT\" NAME=\"def_txcolor\"", MSG(M_MODTXCOLOR));
      printf(" VALUE=\"%s\" SIZE=10><BR>\n", def_txcolor);
    printf("%s<BR>\n", MSG(M_MODLICOLOR));
    printf("<NOBR>LINK:<INPUT TYPE=\"TEXT\" NAME=\"def_licolor\"");
      printf(" VALUE=\"%s\" SIZE=\"10\" MAXLENGTH=\"%d\"></NOBR>\n",
        def_licolor, COLOR_LEN - 1);
    printf("<NOBR>ALINK:<INPUT TYPE=\"TEXT\" NAME=\"def_alcolor\"");
      printf(" VALUE=\"%s\" SIZE=\"10\" MAXLENGTH=\"%d\"></NOBR>\n",
        def_alcolor, COLOR_LEN - 1);
    printf("<NOBR>VLINK:<INPUT TYPE=\"TEXT\" NAME=\"def_vlcolor\"");
      printf(" VALUE=\"%s\" SIZE=\"10\" MAXLENGTH=\"%d\"><NOBR><BR>\n",
        def_vlcolor, COLOR_LEN - 1);
    printf("%s:<INPUT TYPE=\"TEXT\" NAME=\"def_dncolor\"", MSG(M_MODDNCOLOR));
      printf(" VALUE=\"%s\" SIZE=\"10\" MAXLENGTH=\"%d\">\n",
        def_dncolor, COLOR_LEN - 1);
    printf("%s:<INPUT TYPE=\"TEXT\" NAME=\"def_dtcolor\"", MSG(M_MODDTCOLOR));
      printf(" VALUE=\"%s\" SIZE=\"10\" MAXLENGTH=\"%d\"><BR>\n",
        def_dtcolor, COLOR_LEN - 1);
    printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_MODIFY));
    printf("<INPUT TYPE=\"RESET\" VALUE=\"%s\">\n", MSG(M_UNDO));
    print_common("customize");
    printf("</FORM>\n");

    printf("<HR>\n");
    printf("<H3>%s</H3>\n", MSG(M_LINK));

    printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"_top\">\n",
      cgimtd, cgibin);
    for(i=0;i<MAX_LINKS;i++)
    {
      printf("URL:<INPUT TYPE=\"TEXT\" NAME=\"link_url\"");
        printf(" VALUE=\"%s\" SIZE=\"56\" MAXLENGTH=\"%d\"><BR>\n",
          chat->links[i].url, ADDR_LEN - 1);
      printf(">>>>>><INPUT TYPE=\"TEXT\" NAME=\"link_label\"");
        printf(" VALUE=\"%s\" SIZE=\"32\" MAXLENGTH=\"%d\"><BR>\n",
          chat->links[i].label_ja, LABEL_LEN - 1);
    }
    printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_MODIFY));
    printf("<INPUT TYPE=\"RESET\" VALUE=\"%s\">\n", MSG(M_UNDO));
    print_common("customize");
    printf("</FORM>\n");

    printf("<HR>\n");
    printf("<FORM METHOD=\"%s\" ACTION=\"%s\" TARGET=\"_top\">\n",
      cgimtd, cgibin);
    printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_INITIALIZE));
    print_common("reset");
    printf("</FORM>\n");
    printf("<HR>\n");
#ifdef EXPORT_INI
    if(access(EXPORT_INI, W_OK) == -1)
    {
      printf(MSG(M_CANNOTWRITE), EXPORT_INI);
      printf("<BR>\n");
    }
#endif
#ifdef EXPORT_USR
    if(access(EXPORT_USR, W_OK) == -1)
    {
      printf(iMSG(M_CANNOTWRITE), EXPORT_USR);
      printf("<BR>\n");
    }
#endif
#ifdef EXPORT_BBS
    if(access(EXPORT_BBS, W_OK) == -1)
    {
      printf(MSG(M_CANNOTWRITE), EXPORT_BBS);
      printf("<BR>\n");
    }
#endif
#ifdef EXPORT_CHAT
    if(access(EXPORT_CHAT, W_OK) == -1)
    {
      printf(MSG(M_CANNOTWRITE), EXPORT_CHAT);
      printf("<BR>\n");
    }
#endif
  }

  printf("</BODY>\n");
  printf("</HTML>\n");
  chat_exit(0);
}

void printlinks(chat)
CHAT *chat;
{
  int i;
  printf("<HR>\n");
  printf("<H3>%s</H3>\n", MSG(M_LINK));
  printf("<UL>\n");
  for(i=0;i<MAX_LINKS;i++)
  {
    if(*chat->links[i].label_ja != '\0')
    {
      printf("<LI><A HREF=\"%s\" TARGET=\"_blank\">", chat->links[i].url);
      htmlout(chat->links[i].label_ja, 1);
      printf("</A><BR>\n");
    }
  }
  printf("</UL>\n");
}


void printdebug(chat)
CHAT *chat;
{
  int i, j;
  int users[MAX_USERS];

  /* sort by atime */
  for(i=0;i<MAX_USERS;i++) users[i] = i;
  for(i=0;i<MAX_USERS-1;i++)
  {
    for(j=i+1;j<MAX_USERS;j++)
    {
      int k;
      if(chat->user[users[i]].atime < chat->user[users[j]].atime)
      {
        k = users[j];
        users[j] = users[i];
        users[i] = k;
      }
    }
  }

  printf("Content-type: text/plain\n\n");
  printf("memory size: %d version %2.2f\n", sizeof(CHAT), memversion/100.0);
  for(i=0;i<MAX_USERS;i++)
  {
    int addrlen;
    j = users[i];
    if(*chat->user[j].password != '\0')
      printf("*");
    else
      printf(" ");
    if(chat->user[j].permit > 0)
      printf("+");
    else
      printf(" ");
    if(chat->user[j].permit > 1)
      printf("+");
    else
      printf(" ");
    printf(" %2d ", users[i]);
    printf("["); printdate(NULL, chat->user[j].logintime); printf(" - ");
    printdate(NULL, chat->user[j].atime); printf("]");
    printf(" [%s:%d]", chat->user[j].name, chat->user[j].mid);
    printf(" [%s] [%s] [%d:%d]", chat->user[j].address, chat->user[j].agent,
        chat->user[j].login, chat->user[j].count);
    printf("["); i2ip(NULL, chat->user[j].haddressip);
    printf(":%s]", chat->user[j].haddress);
    printf("[%s]", chat->user[j].referer);
    addrlen = strlen(chat->user[j].address);
    if(chat->user[j].address[addrlen+1] == '#')
    {
      printf("[%s]", &chat->user[j].address[addrlen+2]);
    }
    printf("\n");
  }
  printf("\n");
  for(i=0;i<MAX_DENYHOSTS;i++)
  {
    i2ip(NULL, chat->denyhosts[i].address);
    printf(" ["); printdate(NULL, chat->denyhosts[i].atime); printf("]");
    if(*chat->denyhosts[i].name != '\0')
      printf(" [%s]", chat->denyhosts[i].name);
    printf("\n");
  }
  for(i=0;i<MAX_LINKS;i++)
  {
    printf(" [%s][%s][%s]\n", chat->links[i].url, chat->links[i].label_ja,
      chat->links[i].label_en);
  }
}

void printdate(s, t)
char *s;
time_t t;
{
  struct tm *tim;
  time_t lt;
  char *mon[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
    "Sep", "Oct", "Nov", "Dec"};
  lt = t + SHIFT_TIME;
  tim = localtime(&lt);
  if(s != NULL)
  {
    if(strcmp(lang, "en") == 0)
    {
      sprintf(s, "%s %d, %02d:%02d%s",
        mon[tim->tm_mon], tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
    else
    {
      sprintf(s, "%d/%d %02d:%02d%s",
        tim->tm_mon + 1, tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
  }
  else
  {
    if(strcmp(lang, "en") == 0)
    {
      printf("%s %d, %02d:%02d%s",
        mon[tim->tm_mon], tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
    else
    {
      printf("%d/%d %02d:%02d%s",
        tim->tm_mon + 1, tim->tm_mday, tim->tm_hour, tim->tm_min, TZ_STRING);
    }
  }
}

void printchat(chat, lines)
CHAT *chat;
int lines;
{
  int i, j, t = 0;

  if(userreseted)
  {
    printf("<SCRIPT LANGUAGE=\"JavaScript\">\n");
    printf("  top.location = \"%s\";\n", cgibin);
    printf("</SCRIPT>\n");
    return;
  }

  if(*frame != '\0')
  {
    printf("<HR>\n");
    noframe_input();
  }
  
  if(chat->topic.atime >= now - 60)
  {
    printtopic(chat);
    t = 1;
  }

  printf("<HR>\n");
  /* print chattext */
  for(i = j = 0; i < MAX_LINES; i++)
  {
    int pline;
    pline = (chat->text_line - 1 - i + MAX_LINES)%MAX_LINES;
    if((chat->text[pline].atime > 0) &&
       (*chat->text[pline].message != '\0') &&
       (chat->text[pline].room == room) )
    {
      if(*chat->text[pline].namecolor != '\0')
        printf("<FONT COLOR=\"%s\">", chat->text[pline].namecolor);
      else
        printf("<FONT COLOR=\"%s\">", def_dncolor);
      if(chat->text[pline].flags & TEXT_ITALIC) printf("<I>");
      htmlout(chat->text[pline].name, 1);
      if(chat->text[pline].flags & TEXT_ITALIC) printf("</I>");
      printf("&gt; </FONT>");
      if(*chat->text[pline].textcolor != '\0')
        printf("<FONT COLOR=\"%s\">", chat->text[pline].textcolor);
      htmlout(chat->text[pline].message, 1);
      if(*chat->text[pline].textcolor != '\0')
        printf("</FONT>");
      printf("  <FONT COLOR=\"%s\" SIZE=\"-1\">(", def_dtcolor);
      printdate(NULL, chat->text[pline].atime);
      printf(")</FONT><BR>\n");
#ifdef TEXT_HR
      printf("<HR>");
#endif
      if(++j >= lines) break;
    }
  }
  if(t == 0)
    printtopic(chat);
}

void printtopic(chat)
CHAT *chat;
{
  if(chat != NULL)
  {
    /* print topic */
    if((chat->topic.atime > 0) &&
       (*chat->topic.message != '\0'))
    {
      printf("<HR>\n");
      printf("%s: ", MSG(M_TOPIC));
      htmlout(chat->topic.message, 1);
      printf(" (");
      printdate(NULL, chat->topic.atime);
      printf(" "); htmlout(chat->topic.name, 1); printf(")\n");
    }
  }
}

void printbbs(chat, lines)
CHAT *chat;
int lines;
{
  int i, m;
  char pname[256];
  char paddr[256];

  m = ismanager();
  if(sjis)
  {
    strncpy(pname, name, sizeof(pname));
    strncpy(paddr, address, sizeof(paddr));
  }
  else
  {
    sjis2euc(pname, name);
    sjis2euc(paddr, address);
  }
  print_content();
  printf("<TITLE>"); nohtmlout(title);
  printf("(%s)</TITLE></HEAD>\n", MSG(M_BBS));
  body_out(&body);
  printf("<H3>"); htmlout(title, 1);
  printf("(%s)</H3>\n", MSG(M_BBS));
  printf("%s: ", MSG(M_NAME)); htmlout(name, 1); printf("<BR>\n");

  if(permission > 0)
  {
    printf("<FORM METHOD=\"%s\" ACTION=\"%s\">\n", cgimtd, cgibin);
    printf("<TEXTAREA NAME=\"chat\" ROWS=\"3\" COLS=\"60\">");
    printf("%s", message);
    printf("</TEXTAREA><BR>\n");
    printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_BBSWRITE));
    printf("<INPUT TYPE=\"RESET\" VALUE=\"%s\">\n", MSG(M_CLEAR));
    printf("<INPUT TYPE=\"CHECKBOX\" NAME=\"secure\">%s\n", MSG(M_SECURE));
    print_common("bbs");
    printf("</FORM>\n");
  }

  /* print bbs */
  for(i = 0; i < lines; i++)
  {
    int pline;
    pline = (chat->bbs_line - 1 - i + MAX_BBS)%MAX_BBS;
    if((chat->bbs[pline].atime > 0) &&
       ( *chat->bbs[pline].message != '\0'))
    {
      int sec = 0;
      if((chat->bbs[pline].flags & SECURE_BBS) != 0)
        sec = 1;
      if(permission == 0 && sec == 1) continue;
      printf("<HR>\n");
      printf("<FONT SIZE=4 COLOR=\"%s\">", def_dncolor);
      if(sec)
        printf("*** ");
      printdate(NULL, chat->bbs[pline].atime);
      printf(" (");
      htmlout(chat->bbs[pline].name, 1);
      printf(")");
      if(sec)
        printf(" ***");
      printf("</FONT><BR>\n");
      printf("<PRE>\n");
      htmlout(chat->bbs[pline].message, 1);
      printf("</PRE>\n");
      if(strcmp(name, chat->bbs[pline].name) == 0 || m)
      {
        printf("<FORM METHOD=\"%s\" ACTION=\"%s\">\n", cgimtd, cgibin);
        printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"%s\">\n", MSG(M_DELETE));
        print_common("bbs");
        printf("<INPUT TYPE=\"HIDDEN\" NAME=\"chat\"");
        printf(" VALUE=\"<delete %d>\">\n", pline);
        printf("</FORM>\n");
      }
    }
  }
  footer();
}

void sigalm()
{
  signal(SIGALRM, SIG_IGN);
  if(outhead == 0)
    header(NULL, 0);
  printf("Timeout!\n");
  if(chat != NULL)
  {
    unlock_chat(chat);
    release_chat(chat);
  }
  footer();
  exit(1);
}

void sigterm(sig)
int sig;
{
  signal(SIGSEGV, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGINT, SIG_IGN);
  signal(SIGPIPE, SIG_IGN);
  if(outhead == 0)
    header(NULL, 0);
  printf("Terminated!!(%d)\n", sig);
  if(chat != NULL)
  {
    unlock_chat(chat);
    release_chat(chat);
  }
  footer();
  exit(1);
}

void killtag(str)
char *str;
{
  int i;
  for(i=0;str[i] != '\0';i++)
    if(str[i] == '<' || str[i] == '>') str[i] = ' ';
  for(i=strlen(str)-1;i>=0;i--)
    if(str[i] > 0 && str[i] <= ' ') str[i] = '\0';
}

char *getcgiarg()
{
  int param_len;
  int postmethod = 0;
  char *param;
  static char *null = "";


  if(getenv("REQUEST_METHOD") != NULL)
  {
    if(strcmp((char *)getenv("REQUEST_METHOD"), "POST") == NULL)
      postmethod = 1;
  }
  else
    return(null);

  if(postmethod == 0)
  {
    char *p;
    if((p = getenv("QUERY_STRING")) != NULL)
    {
      if((param = malloc(strlen(p)+1)) != NULL)
        strcpy(param, p);
      else
        param = null;
    }
    else
      param = null;
  }
  else
  {
    if(getenv("CONTENT_LENGTH") != NULL)
    {
      param_len = atoi(getenv("CONTENT_LENGTH"));
      if((param = malloc(param_len+1)) != NULL)
        fgets(param, param_len+1, stdin);
      else
        param = null;
    }
    else
      param = null;
  }
  return(param);
}

int get_userentry()
{
  int i;
  
  if(*password != '\0') /* with a password */
  {
    for(i=0;i<MAX_USERS;i++)
    {
      if(strcmp(chat->user[i].name, name) == 0)
      {
        if(*chat->user[i].password == '\0')
          return(i);
        if(strcmp(chat->user[i].password, password) == 0)
          return(i);
        /* invalid password */
        *name = '\0';
        return(-1);
      }
    }
  }
  else /* null password */
  {
    for(i=0;i<MAX_USERS;i++)
    {
      if(strcmp(chat->user[i].name, name) == 0)
      {
        if(*chat->user[i].password != '\0')
        {
          /* invalid password */
          *name = '\0';
          return(-1);
        }
        if(compareIP(chat->user[i].haddressip, haddressip, 0xffffff00) == 0)
          return(i);
      }
    }
  }
  return(-1);
}

int new_userentry()
{
  int i;
  int older;

  // find null entry
  for(i=0;i<MAX_USERS;i++)
  {
    if((chat->user[i].atime == 0) && (chat->user[i].permit == 0) &&
       (*chat->user[i].password == '\0'))
      return(i);
  }
  older = 0;
  for(i=1;i<MAX_USERS;i++)
  {
    if((chat->user[older].atime > chat->user[i].atime) &&
       (chat->user[i].permit <= DEFAULT_PERMIT) &&
       (*chat->user[i].password == '\0'))
      older = i;
  }
  if(chat->user[older].atime > now - 3600)
  {
    for(i=1;i<MAX_USERS;i++)
    {
      if((chat->user[older].atime > chat->user[i].atime) &&
         (chat->user[i].permit <= DEFAULT_PERMIT))
        older = i;
    }
  }
  return(older);
}

int setuserinfo()
{
  int i;
  int newuser = 0;

  if((i = get_userentry()) == -1)
  { /* new user */
    if(*name == '\0')
    {
      sendmessage(alertmsg, sizeof(alertmsg), MSG(M_INVALIDPASS), MT_ERROR);
      back = 1;
      scerr(MSG(M_INVALIDPASS));
    }
    if((i = new_userentry()) == -1)
    {
      scerr(MSG(M_CANNOTUSE));
    }
    chat->user[i].count = 0;
    chat->user[i].refresh = 0;
    chat->user[i].permit = DEFAULT_PERMIT;
    *chat->user[i].memo = '\0';
    *chat->user[i].address = '\0';
    chat->user[i].maxlines = DEF_MAX_LINES;
    *chat->user[i].namecolor = '\0';
    *chat->user[i].textcolor = '\0';
    newuser = 1;
  }
  strncpy(chat->user[i].name, name, sizeof(name));
  strncpy(chat->user[i].password, password, sizeof(password));
  strncpy(chat->user[i].agent, agent, sizeof(agent));
  strncpy(chat->user[i].haddress, haddress, sizeof(haddress));
  chat->user[i].haddressip = haddressip;
  if(autoreload == 0 || chat->user[i].atime < now - AUTOLOGOUT*2)
    chat->user[i].ctime = now;
  chat->user[i].atime = now;
  chat->user[i].room = room;
  switch(chat_mode)
  {
  case m_logout:
    chat->user[i].login = 0;
    break;
  case m_entry:
    {
      int addrlen;
      chat->user[i].logintime = now;
      strncpy(chat->user[i].address, address, sizeof(address));
      strncpy(chat->user[i].referer, referer, sizeof(referer));
      if(CONV == 1)
        chat->user[i].flags |= USER_ITALIC;
      else
        chat->user[i].flags &= ~USER_ITALIC;
      chat->user[i].refresh = refresh;
      addrlen = strlen(chat->user[i].address);
#ifdef COOKIE
      if(addrlen < ADDR_LEN - 2)
      {
        if(*cookie != '\0')
        {
          chat->user[i].address[addrlen+1] = '#';
          strncpy(&chat->user[i].address[addrlen+2], &cookie[5], ADDR_LEN - addrlen - 2);
        }
        else
          chat->user[i].address[addrlen+1] = '\0';
      }
#endif /* COOKIE */
    }
  default:
    chat->user[i].login = 1;
    maxlines = chat->user[i].maxlines;
    if(*chat->user[i].namecolor != '\0')
      strncpy(namecolor, chat->user[i].namecolor, COLOR_LEN);
    if(*chat->user[i].textcolor != '\0')
      strncpy(textcolor, chat->user[i].textcolor, COLOR_LEN);
  }
  if(ismanager())
    chat->user[i].permit = 2;
  else if(newuser && *password != '\0')
    export_user(EXPORT_USER, &chat->user[i]);

  return(i);
}

char *getparam(dd, maxlen, src)
char *dd;
int maxlen;
char *src;
{
  char d[MSG_SIZE], *dst = d, *p;
  int i = 0, jis = 0;

  while(1)
  {
    if(*src == '&' || *src == 0x0d || *src == 0x0a || *src == 0) break;
    if(*src == '%')
    {
      int c;
      c = (h2i(*(src+1))<<4) | h2i(*(src+2));
      dst[i++] = c;
      src+=3;
    }
    else if(*src == '+')
    {
      dst[i++] = ' ';
      src++;
    }
    else
      dst[i++] = *src++;
    if(i >= maxlen - 1)
    {
       while((*src != '&') && (*src != 0x0d) &&
             (*src != 0x0a) && (*src != 0)) src++;
       break;
    }
  }
  dst[i] = '\0';
  while(*src == 0x0d || *src == 0x0a) src++;
  if(*src == '&') src++;
  if((p = (char *)strchr(d, ESC)) != NULL)
  {
    if(strlen(p)>3 && *(p+1) == '$') jis = 1;
  }
  
  if(sjis && jis == 0)
    strncpy(dd, d, maxlen);
  else
    euc2sjis(dd, d);
  return(src);
}

void paramencode(dst, src)
char *dst;
char *src;
{
  int i;
  for(i=0;src[i] != '\0';i++)
  {
    if(src[i] == '%' || src[i] == '+' || src[i] == '#' || src[i] == '"' || src[i] == '&' ||
       (src[i]>=0 && src[i]<=' '))
    {
      *dst++ = '%';
      *dst++ = i2h((src[i]>>4)&0x0f);
      *dst++ = i2h(src[i]&0x0f);
    }
    else if((src[i]&0x80) != 0) /* kanji */
    {
      *dst++ = '%';
      *dst++ = i2h((src[i]>>4)&0x0f);
      *dst++ = i2h(src[i]&0x0f);
      i++;
      *dst++ = '%';
      *dst++ = i2h((src[i]>>4)&0x0f);
      *dst++ = i2h(src[i]&0x0f);
    }
    else
      *dst++ = src[i];
  }
  *dst = '\0';
}

void show_users(chat, showaddr)
CHAT *chat;
int showaddr;
{
  int i, j;
  int u, users[MAX_USERS];

  /* get active users */
  for(u=0,i=0;i<MAX_USERS;i++)
  {
    if(chat->user[i].atime > now - AUTOLOGOUT &&
        *chat->user[i].name != '\0')
    {
      if((chat->user[i].login == 0) && chat->user[i].atime < now - 30) continue;
      users[u++] = i;
    }
  }

  /* sort by count */
  for(i=0;i<u-1;i++)
  {
    for(j=i+1;j<u;j++)
    {
      int k;
      if(chat->user[users[i]].count < chat->user[users[j]].count)
      {
        k = users[j];
        users[j] = users[i];
        users[i] = k;
      }
    }
  }


  for(j=0;j<u;j++)
  {
    int i;
    int blink = 0;
    int hlink = 0;
    int smalluser = 0;
    i = users[j];

    if(userproperty)
      printf("<BR>\n<NOBR>");

    if(room != chat->user[i].room)
    {
      smalluser = 1;
      printf("<FONT SIZE=-1>");
    }
    printf(" [");
    if(chat->user[i].atime < now - BLINKTIME || chat->user[i].login != 1)
    {
      printf("<BLINK>");
      blink = 1;
    }
    if((*chat->user[i].address != '\0') && (showaddr == 1) &&
        (chat->user[i].login == 1))
    {
      if(strchr(chat->user[i].address, ':') != NULL)
        printf("<A HREF=\"%s\">", chat->user[i].address);
      else
        printf("<A HREF=\"mailto:%s\" TARGET=\"_top\">", chat->user[i].address);
      if(chat->user[i].flags & USER_ITALIC) printf("<I>");
      htmlout(chat->user[i].name, 1);
      if(chat->user[i].flags & USER_ITALIC) printf("</I>");
      hlink = 1;
    }
    else
    {
      if(chat->user[i].flags & USER_ITALIC) printf("<I>");
      htmlout(chat->user[i].name, 1);
      if(chat->user[i].flags & USER_ITALIC) printf("</I>");
    }
    if(chat->user[i].login == 0)
      printf("(bye!)");
    else if(chat->user[i].login == 2)
      printf("(dead!)");
    if(chat->user[i].ctime < now - SLEEPTIME && chat->user[i].login == 1)
    {
      int h;
      h = (now - chat->user[i].ctime)/3600;
      printf("(Zzz");
      if(h > 0)
        printf("%d", h);
      printf(")");
    }
    if(hlink) printf("</A>");
    if(blink) printf("</BLINK>");
    printf("]");
    if(smalluser) printf("</FONT>");

    if(userproperty)
    {
      int addrlen = strlen(chat->user[i].address);
      int ref = 0;
      if(*chat->user[i].password != '\0')
        printf("*");
      if(chat->user[i].permit)
        printf("+");
      if(permission >= 2)
        if(chat->user[i].permit == 2)
          printf("+");
      printf(":&lt;%d&gt;:", chat->user[i].count);
      if(*chat->user[i].referer != '\0')
      {
        printf("<a href=\"%s\">", chat->user[i].referer);
        ref = 1;
      }
      i2ip(NULL, chat->user[i].haddressip);
      if(ref) printf("</a>");
      printf("(%s):", chat->user[i].haddress);
      printf("%s", chat->user[i].agent);
#ifdef COOKIE
      if((permission >= 2) && (chat->user[i].address[addrlen+1] == '#'))
      {
        printf("[%s]", &chat->user[i].address[addrlen+2]);
      }
#endif /* COOKIE */
      printf("</NOBR>");
    }

  }
  printf("\n");
}


void header(chat, showaddr)
CHAT *chat;
int showaddr;
{
  int refresh = 0;
  if(userid != -1 && chat_mode == m_chattext) refresh = chat->user[userid].refresh;
  print_content();
  printf("<TITLE>"); nohtmlout(title); printf("</TITLE>\n");
  if(refresh > 0)
  {
    char pname[256];
    char ename[256];
    char pbgcolor[41];
    char ptxcolor[41];
    if(sjis)
      strncpy(ename, name, sizeof(ename));
    else
      sjis2euc(ename, name);
    paramencode(pname, ename);
    paramencode(pbgcolor, bgcolor);
    paramencode(ptxcolor, txcolor);
    printf("<META HTTP-EQUIV=\"Refresh\" CONTENT=\"%d;", refresh);
    printf("URL=%s?auto&", cgibin);
    printf("mode=chattext");
    if(room != 0)
      printf("&room=%d", room);
    if(*name != '\0')
      printf("&name=%s", pname);
    if(*password != '\0')
      printf("&password=%s", cpassword);
    if(*bgcolor != '\0')
      printf("&bgcolor=%s", pbgcolor);
    if(*txcolor != '\0')
      printf("&txcolor=%s", ptxcolor);
    if(*lang != '\0')
      printf("&lang=%s", lang);
    printf("\">\n");
  }
  if(chat_mode == m_chattext)
    printf("<BASE TARGET=\"_blank\">\n");
  printf("</HEAD>\n");
  alertmessages();
  body_out(&body);
  if(hasalert)
  {
    printf("<SCRIPT LANGUAGE=\"JavaScript\">\n");
    printf("showalert()\n");
    printf("</SCRIPT>\n");
    if(alertbtn)
    {
      printf("<FORM><INPUT TYPE=\"button\" VALUE=\"alert!\" ONCLICK=\"showalert()\">\n");
      printf("</FORM>\n");
    }
  }
  if(back) historyback();
  printf("<FONT SIZE=\"4\"><STRONG>");
  htmlout(title, 1);
  printf("</STRONG></FONT>");
  if(refresh > 0 && showaddr)
  {
    printf("<FONT COLOR=\"%s\"> (refresh=%dsec.", def_dtcolor, refresh);
    if(room != 0) printf("/room=%d", room);
    if(userid >= 0 && *password != '\0')
    { /* count unread bbs */
      int urcnt = 0, i;
      for(i=0;i<MAX_BBS;i++)
      {
        int lin = (chat->bbs_line-1-i+MAX_BBS)%MAX_BBS;
        if(chat->bbs[lin].atime == 0) continue;
        if( (chat->user[userid].btime < chat->bbs[lin].atime) &&
            !(((chat->bbs[lin].flags & SECURE_BBS) != 0) && (permission == 0)) )
          urcnt ++;
      }
      if(urcnt > 0)
        printf("/%s%d\n", MSG(M_UNREADBBS), urcnt);
    }
    printf(")</FONT>\n");
  }
  printf("<P>\n");
  if(chat == NULL) return;
  printf("%s: ", MSG(M_MEMBERS));
  show_users(chat, showaddr);

  outhead = 1;
}

void footer()
{
  printf("<HR>\n");
  printf("<P ALIGN=\"right\">");
  printf("<A HREF=\"%s\">", official_url);
  printf("%s ver%2.2f", MSG(M_SOFTWARE), (double)version/100);
  if(patchlevel > 0) printf("%c", '\x60' + (char)patchlevel);
  if(patchlevel == -1) printf("beta");
  printf("</A></P>\n");
  printf("</BODY>\n");
  printf("</HTML>\n");
}

void scerr(msg)
char *msg;
{
  header(NULL, 0);
  printf("<H1>%s</H1>\n", msg);
  footer();
  if(chat != NULL)
  {
    unlock_chat(chat);
    release_chat(chat);
  }
  exit(1);
}

